/*
 * Copyright (c) 2010 ICM Uniwersytet Warszawski All rights reserved. See LICENCE file for licensing
 * information.
 * 
 * Author: T. Rękawek <newton@mat.uni.torun.pl>
 */

package de.fzj.unicore.uas.dsms;

import static org.junit.Assert.fail;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.security.auth.x500.X500Principal;

import junit.framework.Assert;

import org.apache.commons.io.FileUtils;
import org.oasisOpen.docs.wsrf.sg2.EntryType;
import org.unigrids.services.atomic.types.GridFileType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import pl.edu.icm.dsms.catalogue.Catalogue;
import pl.edu.icm.dsms.catalogue.CatalogueClient;
import pl.edu.icm.dsms.catalogue.CatalogueUtils;
import pl.edu.icm.dsms.catalogue.InitCatalogue;
import pl.edu.icm.dsms.catalogue.model.Lda;
import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.client.BaseUASClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.StorageFactoryClient;
import de.fzj.unicore.uas.client.TSFClient;
import de.fzj.unicore.uas.impl.dsms.CreateDSMSOnStartup;
import de.fzj.unicore.uas.impl.dsms.DSMSImpl;
import de.fzj.unicore.uas.impl.registry.RegistryHandler;
import de.fzj.unicore.uas.impl.sms.StorageFactoryHomeImpl;
import de.fzj.unicore.uas.impl.tss.TargetSystemFactoryHomeImpl;
import de.fzj.unicore.uas.impl.tss.TargetSystemFactoryImpl;
import de.fzj.unicore.uas.util.CreateSMSOnStartup;
import de.fzj.unicore.uas.xnjs.XNJSFacade;
import de.fzj.unicore.wsrflite.Home;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.exceptions.ResourceNotCreatedException;
import de.fzj.unicore.wsrflite.security.ISecurityProperties;
import de.fzj.unicore.wsrflite.security.UASSecurityProperties;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import de.fzj.unicore.wsrflite.xmlbeans.impl.WSResourceImpl;
import eu.unicore.security.util.client.ETDClientSettings;
import eu.unicore.security.util.client.HttpUtils;

/**
 * Some utils helpful for testing catalogue.
 * 
 * @author newton
 * 
 */
public abstract class AbstractTestBase {
	private static final String UAS_CONFIG_FILE = "src/test/resources/dcatalogue/uas.config";

	private static final String DEFAULT_CAT_NAME = "default_catalogue";

	protected static final String INVALID_DSMS_ID = "123";

	private static final String CLIENT_KEYSTORE = "src/test/resources/dcatalogue/keystore.jks";

	private static final String CLIENT_KEYSTORE_PASS = "the!njs";

	protected static final String[] CLIENT_ALIASES = new String[] { "client1", "client2", "client3" };

	protected static final String DEFAULT_CLIENT_ALIAS = CLIENT_ALIASES[0];

	private Home factoryHome, smsHome, catalogueHome, dSmsHome, tsfHome;

	private String smsResId, factoryResId, dSmsResId, tsfResId;

	private InitCatalogue initCatalogue;

	protected static Kernel kernel;

	protected AbstractTestBase() throws Exception {
		// setSecurityProperties(UAS_CONFIG_FILE);
	}

	protected static void startKernel() throws Exception {
		FileUtils.deleteQuietly(new File("target", "data"));
		System.setProperty("java.util.logging.config.file", "src/test/resources/dcatalogue/log4j.properties");
		kernel = new Kernel();
		kernel.startSynchronous(UAS_CONFIG_FILE);
		RegistryHandler.get(kernel);
		// start XNJS and link kernel instance to XNJS instance
		XNJSFacade.get(null, kernel);
	}

	protected static void shutdownKernel() {
		// kernel.shutdown();
	}

	public static void cleanStorageDirs() {
		File testsRoot = new File("target", "unicorex-test");
		File s1 = new File(testsRoot, "storagefactory");
		File s2 = new File(testsRoot, "teststorage");
		FileUtils.deleteQuietly(s1);
		FileUtils.deleteQuietly(s2);
		s1.mkdir();
		s2.mkdir();
	}

	protected EndpointReferenceType findFirstAccessibleService(List<EndpointReferenceType> eprs) {
		for (EndpointReferenceType epr : eprs) {
			try {
				BaseUASClient c = new BaseUASClient(epr, kernel.getSecurityProperties());
				c.getCurrentTime();
				return epr;
			} catch (Exception e) {
			}
		}
		return null;
	}

	protected boolean compareEprList(List<EndpointReferenceType> list1, List<Lda> list2) {
		if (list1.size() != list2.size())
			return false;

		for (EndpointReferenceType epr1 : list1) {
			boolean ok = false;
			for (Lda lda : list2)
				if (compareEpr(epr1, lda.getSms())) {
					ok = true;
					break;
				}
			if (!ok)
				return false;
		}
		return true;
	}

	protected boolean compareFileList(GridFileType[] list1, String[] list2) {
		if (list1.length != list2.length) {
			return false;
		}

		for (GridFileType file1 : list1) {
			boolean ok = false;
			for (String file2 : list2)
				if (file1.getPath().equals(file2)) {
					ok = true;
					break;
				}
			if (!ok) {
				return false;
			}
		}
		return true;
	}

	protected boolean compareEpr(EndpointReferenceType sms1, EndpointReferenceType sms2) {
		return sms1.getAddress().getStringValue().equals(sms2.getAddress().getStringValue());
	}

	protected RegistryClient getRegistryClient() throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + UAS.REG + "?res=default_registry");

		return new RegistryClient(epr, getSecurityProperties());
	}

	protected CatalogueClient getCatalogueClient() throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url + "/" + Catalogue.CATALOGUE + "?res=" + DEFAULT_CAT_NAME);

		return new CatalogueClient(epr, getSecurityProperties());
	}

	protected StorageClient getSmsClient() throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL) + "/" + UAS.SMS + "?res="
				+ CreateSMSOnStartup.INSTANCE_ID;
		return getStorageClient(url, DEFAULT_CLIENT_ALIAS);
	}

	protected StorageFactoryClient getFactoryClient() throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL);
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(
				url + "/" + UAS.SMF + "?res=" + StorageFactoryHomeImpl.DEFAULT_SMF_NAME);

		return new StorageFactoryClient(epr, getSecurityProperties());
	}

	protected StorageClient getDsmsClient() throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL) + "/" + UAS.SMS + "?res=default_dsms";
		return getStorageClient(url, DEFAULT_CLIENT_ALIAS);
	}

	protected StorageClient getDsmsClient(String clientAlias) throws Exception {
		String url = kernel.getProperty(Kernel.WSRF_BASEURL) + "/" + UAS.SMS + "?res=default_dsms";
		return getStorageClient(url, clientAlias);
	}

	protected void createSms() throws ResourceNotCreatedException {
		CreateSMSOnStartup r = new CreateSMSOnStartup(kernel);
		r.run();

		smsHome = kernel.getHome(UAS.SMS);
		smsResId = CreateSMSOnStartup.INSTANCE_ID;
		/*
		 * Map<String,Object>map = new HashMap<String,Object>(); map.put(WSResourceImpl.INIT_UNIQUE_ID,
		 * CreateSMSOnStartup.INSTANCE_ID); map.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());
		 * map.put(UASWSResourceImpl.INITPARAM_PUBLISH_TO_REGISTRY, Boolean.TRUE);
		 * map.put(StorageManagementHomeImpl.INIT_STORAGETYPE, StorageTypes.VARIABLE);
		 * map.put(SMSBaseImpl.INIT_FILTERLISTINGS, Boolean.TRUE); map.put(SMSBaseImpl.INIT_STORAGENAME,
		 * "SHARE");
		 * 
		 * String dir = kernel.getProperty(CreateSMSOnStartup.CONF_WORKDIR);
		 * map.put(SMSBaseImpl.INIT_DIRECTORY, dir);
		 * 
		 * smsHome = kernel.getHome(UAS.SMS); smsResId = smsHome.createWSRFServiceInstance(map);
		 */
	}

	protected void destroySms() throws Exception {
		smsHome.destroyWSRFServiceInstance(smsResId);
	}

	protected void createFactory() throws ResourceNotCreatedException {
		String defaultSmfName = StorageFactoryHomeImpl.DEFAULT_SMF_NAME;

		Map<String, Object> imap = new HashMap<String, Object>();
		imap.put(WSResourceImpl.INIT_UNIQUE_ID, defaultSmfName);
		imap.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());

		factoryHome = kernel.getHome(UAS.SMF);
		factoryResId = factoryHome.createWSRFServiceInstance(imap);
	}

	protected void destroyFactory() throws Exception {
		factoryHome.destroyWSRFServiceInstance(factoryResId);
	}

	protected void createCatalogue() throws Exception {
		initCatalogue = new InitCatalogue();
		initCatalogue.setKernel(kernel);
		initCatalogue.run();
		catalogueHome = kernel.getHome(Catalogue.CATALOGUE);
		Thread.sleep(2000); // wait until dcatalogue registers
	}

	protected void destroyCatalogue() throws Exception {
		initCatalogue.stop();
		catalogueHome.destroyWSRFServiceInstance(DEFAULT_CAT_NAME);
	}

	protected void createDsms() throws ResourceNotCreatedException {
		CreateDSMSOnStartup r = new CreateDSMSOnStartup();
		r.setKernel(kernel);
		r.run();
		DSMSImpl.initCache();

		dSmsHome = kernel.getHome(UAS.SMS);
		dSmsResId = CreateDSMSOnStartup.DEFAULT_RESOURCE_ID;
	}

	protected void destroyDsms() throws Exception {
		dSmsHome.destroyWSRFServiceInstance(dSmsResId);
	}
	
	protected void createTsf() throws ResourceNotCreatedException {
		String defaultTsfName = TargetSystemFactoryHomeImpl.DEFAULT_TSF;

		Map<String, Object> imap = new HashMap<String, Object>();
		imap.put(WSResourceImpl.INIT_UNIQUE_ID, defaultTsfName);
		imap.put(WSResourceImpl.INIT_INITIAL_TERMINATION_TIME, lifeTime());
		imap.put(TargetSystemFactoryImpl.INIT_TSF_NAME, "DEMO-SITE");

		tsfHome = kernel.getHome(UAS.TSF);
		tsfResId = tsfHome.createWSRFServiceInstance(imap);
	}

	protected void destroyTsf() throws Exception {
		tsfHome.destroyWSRFServiceInstance(tsfResId);
	}

	protected TSFClient getTsfClient() throws Exception {
		String url = "https://localhost:65321/services/TargetSystemFactoryService?res=default_target_system_factory";
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url);
		return new TSFClient(epr, getSecurityProperties());
	}

	public static Calendar lifeTime() {
		Calendar c = new GregorianCalendar();
		c.add(Calendar.MINUTE, 5);
		return c;
	}

	protected static void assertCatEx(CatalogueClient cat, String method, Object[] params,
			Class<?> expectedException) {
		Class<?>[] paramTypes = new Class[params.length];
		try {
			for (int i = 0; i < params.length; i++)
				if (params[i] == null || params[i] instanceof String)
					paramTypes[i] = String.class;
				else if (params[i] instanceof EndpointReferenceType)
					paramTypes[i] = EndpointReferenceType.class;
				else
					paramTypes[i] = params[i].getClass();
			Method m = CatalogueClient.class.getMethod(method, paramTypes);
			m.invoke(cat, params);
		} catch (NoSuchMethodException e) {
			StringBuilder builder = new StringBuilder("No such method: ");
			builder.append(method).append('(');
			for (int i = 0; i < paramTypes.length; i++) {
				builder.append(paramTypes[i].getSimpleName());
				if (i < paramTypes.length - 1)
					builder.append(", ");
			}
			builder.append(")");
			e.printStackTrace();
			fail(builder.toString());
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			fail(e.toString());
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			fail(e.toString());
		} catch (InvocationTargetException e) {
			if (e.getCause().getClass().equals(expectedException))
				return;
		}
		fail(expectedException.toString() + " should have been thrown.");
	}

	protected StorageClient getStorageClient(String url, String clientAlias) throws Exception {
		EndpointReferenceType epr = EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(url);
		ISecurityProperties securityProps = getSecurityPropertiesWithTd(clientAlias, epr);
		return new StorageClient(epr, securityProps);
	}

	private ISecurityProperties getSecurityProperties() throws Exception {
		return getSecurityPropertiesWithTd(DEFAULT_CLIENT_ALIAS, null);
	}

	protected ISecurityProperties getSecurityPropertiesWithTd(String clientAlias, EndpointReferenceType epr)
			throws Exception {
		Properties properties = new Properties();
		properties.setProperty(ISecurityProperties.WSRF_SSL, "true");
		properties.setProperty(ISecurityProperties.WSRF_SSL_CLIENTAUTH, "true");
		properties.setProperty(ISecurityProperties.WSRF_SSL_KEYALIAS, clientAlias);
		properties.setProperty(ISecurityProperties.WSRF_SSL_KEYSTORE, CLIENT_KEYSTORE);
		properties.setProperty(ISecurityProperties.WSRF_SSL_KEYPASS, CLIENT_KEYSTORE_PASS);

		properties.setProperty(HttpUtils.CONNECT_TIMEOUT, "6000000");
		properties.setProperty(HttpUtils.SO_TIMEOUT, "6000000");

		ISecurityProperties clientProp = new UASSecurityProperties(properties);

		if (epr != null) {
			clientProp.setSignMessage(true);

			String dn = findServerName(epr);
			ETDClientSettings etdSettings = clientProp.getETDSettings();
			etdSettings.setExtendTrustDelegation(true);
			etdSettings.setReceiver(new X500Principal(dn));
		}

		return clientProp;
	}

	protected String findServerName(EndpointReferenceType epr) throws Exception {
		// first, check if server name is already in the EPR...
		String dn = WSUtilities.extractServerIDFromEPR(epr);
		if (dn != null) {
			return dn;
		}
		// otherwise find a matching service in the registry
		String url = epr.getAddress().getStringValue();
		if (url.contains("/services/"))
			url = url.substring(0, url.indexOf("/services"));
		for (EntryType entry : getRegistryClient().listEntries()) {
			if (entry.getMemberServiceEPR().getAddress().getStringValue().startsWith(url)) {
				dn = WSUtilities.extractServerIDFromEPR(entry.getMemberServiceEPR());
				if (dn != null) {
					return dn;
				}
			}
		}
		return null;
	}
	
	protected void assertPathEquals(String expected, String actual) {
		Assert.assertEquals(CatalogueUtils.normalizePath(expected), CatalogueUtils.normalizePath(actual));
	}

}